from datetime import datetime, timedelta, timezone
from typing import Any, Awaitable, Callable, Dict, List, Optional, Union

from pydantic import BaseConfig, BaseModel, validator
from pydantic_openapi_schema.v3_1_0 import (
    Components,
    OAuthFlow,
    OAuthFlows,
    SecurityRequirement,
    SecurityScheme,
)
from starlette.status import HTTP_201_CREATED
from starlite import Cookie, DefineMiddleware, Response
from starlite.enums import MediaType
from starlite.utils import AsyncCallable

from starlite_jwt.middleware import (
    CookieOptions,
    JWTAuthenticationMiddleware,
    JWTCookieAuthenticationMiddleware,
)
from starlite_jwt.token import Token

RetrieveUserHandler = Union[
    Callable[[str], Any],
    Callable[[str], Awaitable[Any]],
]


class JWTAuth(BaseModel):
    """JWT Authentication Configuration.

    This class is the main entry point to the library, and it includes
    methods to create the middleware, provide login functionality, and
    create OpenAPI documentation.
    """

    class Config(BaseConfig):
        arbitrary_types_allowed = True

    algorithm: str = "HS256"
    """
    Algorithm to use for JWT hashing.
    """
    auth_header: str = "Authorization"
    """
    Request header key from which to retrieve the token. E.g. 'Authorization' or 'X-Api-Key'.
    """
    default_token_expiration: timedelta = timedelta(days=1)
    """
    The default value for token expiration.
    """
    retrieve_user_handler: RetrieveUserHandler
    """
    Callable that receives the 'sub' value of a token, which represents the 'subject' of the token (usually a user ID
    or equivalent value) and returns a 'user' value.

    Notes:
    - User can be any arbitrary value,
    - The callable can be sync or async.
    """
    token_secret: str
    """
    Key with which to generate the token hash.

    Notes:
    - This value should be kept as a secret and the standard practice is to inject it into the environment.
    """
    exclude: Optional[Union[str, List[str]]] = None
    """
    A pattern or list of patterns to skip in the authentication middleware.
    """
    openapi_security_scheme_name: str = "BearerToken"
    """
    The value to use for the OpenAPI security scheme and security requirements
    """

    @validator("retrieve_user_handler")
    def validate_retrieve_user_handler(  # pylint: disable=no-self-argument
        cls, value: RetrieveUserHandler
    ) -> Callable[[str], Awaitable[Any]]:
        """This validator ensures that the passed in value does not get bound.

        Args:
            value: A callable fulfilling the RetrieveUserHandler type.

        Returns:
            An instance of AsyncCallable wrapping the callable.
        """
        return AsyncCallable(value)

    @property
    def openapi_components(self) -> Components:
        """Creates OpenAPI documentation for the JWT auth schema used.

        Returns:
            An [Components][pydantic_schema_pydantic.v3_1_0.components.Components] instance.
        """
        return Components(
            securitySchemes={
                self.openapi_security_scheme_name: SecurityScheme(
                    type="http",
                    scheme="Bearer",
                    name=self.auth_header,
                    bearerFormat="JWT",
                    description="JWT api-key authentication and authorization.",
                )
            }
        )

    @property
    def security_requirement(self) -> SecurityRequirement:
        """
        Returns:
            An OpenAPI 3.1 [SecurityRequirement][pydantic_schema_pydantic.v3_1_0.security_requirement.SecurityRequirement] dictionary.
        """
        return {self.openapi_security_scheme_name: []}

    @property
    def middleware(self) -> DefineMiddleware:
        """Creates `JWTAuthenticationMiddleware` wrapped in Starlite's
        `DefineMiddleware`.

        Returns:
            An instance of [DefineMiddleware][starlite.middleware.base.DefineMiddleware].
        """
        return DefineMiddleware(
            JWTAuthenticationMiddleware,
            algorithm=self.algorithm,
            auth_header=self.auth_header,
            retrieve_user_handler=self.retrieve_user_handler,
            token_secret=self.token_secret,
            exclude=self.exclude,
        )

    def login(
        self,
        identifier: str,
        *,
        response_body: Optional[Any] = None,
        response_media_type: Union[str, MediaType] = MediaType.JSON,
        response_status_code: int = HTTP_201_CREATED,
        token_expiration: Optional[timedelta] = None,
        token_issuer: Optional[str] = None,
        token_audience: Optional[str] = None,
        token_unique_jwt_id: Optional[str] = None,
    ) -> Response[Any]:
        """Create a response with a JWT header. Calls the
        'JWTAuth.store_token_handler' to persist the token 'sub'.

        Args:
            identifier: Unique identifier of the token subject. Usually this is a user ID or equivalent kind of value.
            response_body: An optional response body to send.
            response_media_type: An optional 'Content-Type'. Defaults to 'application/json'.
            response_status_code: An optional status code for the response. Defaults to '201 Created'.
            token_expiration: An optional timedelta for the token expiration.
            token_issuer: An optional value of the token 'iss' field.
            token_audience: An optional value for the token 'aud' field.
            token_unique_jwt_id: An optional value for the token 'jti' field.

        Returns:
            A [Response][starlite.response.Response] instance.
        """
        encoded_token = self.create_token(
            identifier=identifier,
            token_expiration=token_expiration,
            token_issuer=token_issuer,
            token_audience=token_audience,
            token_unique_jwt_id=token_unique_jwt_id,
        )
        return Response(
            content=response_body,
            headers={self.auth_header: encoded_token},
            media_type=response_media_type,
            status_code=response_status_code,
        )

    def create_token(
        self,
        identifier: str,
        token_expiration: Optional[timedelta] = None,
        token_issuer: Optional[str] = None,
        token_audience: Optional[str] = None,
        token_unique_jwt_id: Optional[str] = None,
    ) -> str:
        """Creates a Token instance from the passed in parameters, persists and
        returns it.

        Args:
            identifier: Unique identifier of the token subject. Usually this is a user ID or equivalent kind of value.
            token_expiration: An optional timedelta for the token expiration.
            token_issuer: An optional value of the token 'iss' field.
            token_audience: An optional value for the token 'aud' field.
            token_unique_jwt_id: An optional value for the token 'jti' field.

        Returns:
            The created token.
        """
        token = Token(
            sub=identifier,
            exp=(datetime.now(timezone.utc) + (token_expiration or self.default_token_expiration)),
            iss=token_issuer,
            aud=token_audience,
            jti=token_unique_jwt_id,
        )
        encoded_token = token.encode(secret=self.token_secret, algorithm=self.algorithm)

        return encoded_token


class JWTCookieAuth(JWTAuth):
    """JWT Cookie Authentication Configuration.

    This class is an alternate entry point to the library, and it
    includes all of the functionality of the `JWTAuth` class and adds
    support for passing JWT tokens `HttpOnly` cookies.
    """

    auth_cookie: str = "token"
    """
    Cookie name from which to retrieve the token. E.g. 'access-token' or 'refresh-token'.
    """
    auth_cookie_options: "CookieOptions" = CookieOptions(domain=None, secure=False, samesite="lax")
    """
    Cookie configuration options to use when creating cookies for requests
    """

    @property
    def openapi_components(self) -> Components:
        """Creates OpenAPI documentation for the JWT Cookie auth scheme.

        Returns:
            An [Components][pydantic_schema_pydantic.v3_1_0.components.Components] instance.
        """
        return Components(
            securitySchemes={
                self.openapi_security_scheme_name: SecurityScheme(
                    type="http",
                    scheme="Bearer",
                    name=self.auth_cookie,
                    security_scheme_in="cookie",
                    bearerFormat="JWT",
                    description="JWT cookie-based authentication and authorization.",
                )
            }
        )

    @property
    def middleware(self) -> DefineMiddleware:
        """Creates `JWTCookieAuthenticationMiddleware` wrapped in Starlite's
        `DefineMiddleware`.

        Returns:
            An instance of [DefineMiddleware][starlite.middleware.base.DefineMiddleware].
        """
        return DefineMiddleware(
            JWTCookieAuthenticationMiddleware,
            algorithm=self.algorithm,
            auth_header=self.auth_header,
            retrieve_user_handler=self.retrieve_user_handler,
            token_secret=self.token_secret,
            exclude=self.exclude,
            auth_cookie=self.auth_cookie,
            auth_cookie_options=self.auth_cookie_options,
        )

    def login(
        self,
        identifier: str,
        *,
        response_body: Optional[Any] = None,
        response_media_type: Union[str, MediaType] = MediaType.JSON,
        response_status_code: int = HTTP_201_CREATED,
        token_expiration: Optional[timedelta] = None,
        token_issuer: Optional[str] = None,
        token_audience: Optional[str] = None,
        token_unique_jwt_id: Optional[str] = None,
    ) -> Response[Any]:
        """Create a response with a JWT header. Calls the
        'JWTAuth.store_token_handler' to persist the token 'sub'.

        Args:
            identifier: Unique identifier of the token subject. Usually this is a user ID or equivalent kind of value.
            response_body: An optional response body to send.
            response_media_type: An optional 'Content-Type'. Defaults to 'application/json'.
            response_status_code: An optional status code for the response. Defaults to '201 Created'.
            token_expiration: An optional timedelta for the token expiration.
            token_issuer: An optional value of the token 'iss' field.
            token_audience: An optional value for the token 'aud' field.
            token_unique_jwt_id: An optional value for the token 'jti' field.

        Returns:
            A [Response][starlite.response.Response] instance.
        """
        encoded_token = self.create_token(
            identifier=identifier,
            token_expiration=token_expiration,
            token_issuer=token_issuer,
            token_audience=token_audience,
            token_unique_jwt_id=token_unique_jwt_id,
        )
        return Response(
            content=response_body,
            headers={self.auth_header: encoded_token},
            cookies=[
                Cookie(
                    key=self.auth_cookie,
                    value=encoded_token,
                    httponly=True,
                    expires=int(
                        (datetime.now(timezone.utc) + (token_expiration or self.default_token_expiration)).timestamp()
                    ),
                    **self.auth_cookie_options.dict(exclude_none=True),
                )
            ],
            media_type=response_media_type,
            status_code=response_status_code,
        )


class OAuth2PasswordBearerAuth(JWTCookieAuth):
    """OAUTH2 Schema for Password Bearer Authentication.

    This class implements an OAUTH2 authentication flow entry point to the library, and it
    includes all of the functionality of the `JWTAuth` class and adds
    support for passing JWT tokens `HttpOnly` cookies.

    `token_url` is the only additional required argument and should point your login route
    """

    token_url: str
    """
    The URL for retrieving a new token.
    """
    scopes: Optional[Dict[str, str]] = {}
    """Scopes available for the token."""

    @property
    def oauth_flow(self) -> OAuthFlow:
        """Creates an OpenAPI OAuth2 flow for the password bearer
        authentication scheme.

        Returns:
            An [OAuthFlow][pydantic_schema_pydantic.v3_1_0.oauth_flow.OAuthFlow] instance.
        """
        return OAuthFlow(
            tokenUrl=self.token_url,
            scopes=self.scopes,
        )

    @property
    def openapi_components(self) -> Components:
        """Creates OpenAPI documentation for the OAUTH2 Password bearer auth
        scheme.

        Returns:
            An [Components][pydantic_schema_pydantic.v3_1_0.components.Components] instance.
        """
        return Components(
            securitySchemes={
                self.openapi_security_scheme_name: SecurityScheme(
                    type="oauth2",
                    scheme="Bearer",
                    name=self.auth_header,
                    security_scheme_in="header",
                    flows=OAuthFlows(password=self.oauth_flow),  # pyright: reportGeneralTypeIssues=false
                    bearerFormat="JWT",
                    description="OAUTH2 password bearer authentication and authorization.",
                )
            }
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['starlite_jwt']

package_data = \
{'': ['*']}

install_requires = \
['cryptography', 'python-jose', 'starlite>=1.16.0']

setup_kwargs = {
    'name': 'starlite-jwt',
    'version': '1.3.0',
    'description': 'A JWT auth toolkit for Starlite',
    'long_description': '# Starlite JWT\n\n<!-- markdownlint-disable -->\n<img alt="Starlite logo" src="./starlite-banner.svg" width="100%" height="auto">\n<!-- markdownlint-restore -->\n\n<div align="center">\n\n[![Quality Gate Status](https://sonarcloud.io/api/project_badges/measure?project=starlite-api_jwt-auth&metric=alert_status)](https://sonarcloud.io/summary/new_code?id=starlite-api_jwt-auth)\n[![Coverage](https://sonarcloud.io/api/project_badges/measure?project=starlite-api_jwt-auth&metric=coverage)](https://sonarcloud.io/summary/new_code?id=starlite-api_jwt-auth)\n\n[![Maintainability Rating](https://sonarcloud.io/api/project_badges/measure?project=starlite-api_jwt-auth&metric=sqale_rating)](https://sonarcloud.io/summary/new_code?id=starlite-api_jwt-auth)\n[![Security Rating](https://sonarcloud.io/api/project_badges/measure?project=starlite-api_jwt-auth&metric=security_rating)](https://sonarcloud.io/summary/new_code?id=starlite-api_jwt-auth)\n[![Reliability Rating](https://sonarcloud.io/api/project_badges/measure?project=starlite-api_jwt-auth&metric=reliability_rating)](https://sonarcloud.io/summary/new_code?id=starlite-api_jwt-auth)\n[![Code Smells](https://sonarcloud.io/api/project_badges/measure?project=starlite-api_jwt-auth&metric=code_smells)](https://sonarcloud.io/summary/new_code?id=starlite-api_jwt-auth)\n\n[![Discord](https://img.shields.io/discord/919193495116337154?color=blue&label=chat%20on%20discord&logo=discord)](https://discord.gg/X3FJqy8d2j)\n[![Matrix](https://img.shields.io/badge/%5Bm%5D%20chat%20on%20Matrix-bridged-blue)](https://matrix.to/#/#starlitespace:matrix.org)\n\n</div>\n\nThis library offers simple JWT authentication for [Starlite](https://github.com/starlite-api/starlite).\n\nCheckout [the docs 📚](https://starlite-api.github.io/starlite-jwt/).\n\n## Installation\n\n```shell\npip install starlite-jwt\n```\n\nThis library uses the excellent [python-jose](https://github.com/mpdavis/python-jose) library, which supports multiple\ncryptographic backends. You can install either [pyca/cryptography](http://cryptography.io/)\nor [pycryptodome](https://pycryptodome.readthedocs.io/en/latest/), and it will be used as the backend automatically. Note\nthat if you want to use a certificate based encryption scheme, you must install one of these backends - please refer to\nthe [python-jose](https://github.com/mpdavis/python-jose) readme for more details.\n\n## Example\n\n```python\nimport os\nfrom typing import Any, Optional\nfrom uuid import UUID, uuid4\n\nfrom pydantic import BaseModel, EmailStr\nfrom starlite import OpenAPIConfig, Request, Response, Starlite, get\n\nfrom starlite_jwt import JWTAuth, Token\n\n\n# Let\'s assume we have a User model that is a pydantic model.\n# This though is not required - we need some sort of user class -\n# but it can be any arbitrary value, e.g. an SQLAlchemy model, a representation of a MongoDB  etc.\nclass User(BaseModel):\n    id: UUID\n    name: str\n    email: EmailStr\n\n\n# The JWTAuth package requires a handler callable that takes a unique identifier, and returns the \'User\'\n# instance correlating to it.\n#\n# The identifier is the \'sub\' key of the JWT, and it usually correlates to a user ID.\n# It can be though any arbitrary value you decide upon - as long as the handler function provided\n# can receive this value and return the model instance for it.\n#\n# Note: The callable can be either sync or async - both will work.\nasync def retrieve_user_handler(unique_identifier: str) -> Optional[User]:\n    # logic here to retrieve the user instance\n    ...\n\n\n# The minimal configuration required for the library is the callable for the \'retrieve_user_handler\' key, and a string\n# value for the token secret.\n#\n# Important: secrets should never be hardcoded. Its best practice to pass the secret using ENV.\n#\n# Tip: It\'s also a good idea to use the pydantic settings management functionality\njwt_auth = JWTAuth(\n    retrieve_user_handler=retrieve_user_handler,\n    token_secret=os.environ.get("JWT_SECRET", "abcd123"),\n    # we are specifying which endpoints should be excluded from authentication. In this case the login endpoint\n    # and our openAPI docs.\n    exclude=["/login", "/schema"],\n)\n\n\n# Given an instance of \'JWTAuth\' we can create a login handler function:\n@get("/login")\ndef login_handler() -> Response[User]:\n    # we have a user instance - probably by retrieving it from persistence using another lib.\n    # what\'s important for our purposes is to have an identifier:\n    user = User(name="Moishe Zuchmir", email="zuchmir@moishe.com", id=uuid4())\n\n    response = jwt_auth.login(identifier=str(user.id), response_body=user)\n\n    # you can do whatever you want to update the response instance here\n    # e.g. response.set_cookie(...)\n\n    return response\n\n\n# We also have some other routes, for example:\n@get("/some-path")\ndef some_route_handler(request: Request[User, Token]) -> Any:\n    # request.user is set to the instance of user returned by the middleware\n    assert isinstance(request.user, User)\n    # request.auth is the instance of \'starlite_jwt.Token\' created from the data encoded in the auth header\n    assert isinstance(request.auth, Token)\n    # do stuff ...\n\n\n# We add the jwt security schema to the OpenAPI config.\nopenapi_config = OpenAPIConfig(\n    title="My API",\n    version="1.0.0",\n    components=[jwt_auth.openapi_components],\n    security=[jwt_auth.security_requirement],\n)\n\n# We initialize the app instance, passing to it the \'jwt_auth.middleware\' and the \'openapi_config\'.\napp = Starlite(\n    route_handlers=[login_handler, some_route_handler],\n    middleware=[jwt_auth.middleware],\n    openapi_config=openapi_config,\n)\n```\n\n## Contributing\n\nStarlite and all its official libraries is open to contributions big and small.\n\nYou can always [join our discord](https://discord.gg/X3FJqy8d2j) server\nor [join our Matrix](https://matrix.to/#/#starlitespace:matrix.org) space to discuss contributions and project\nmaintenance. For guidelines on how to contribute to this library, please see [the contribution guide](CONTRIBUTING.md).\n',
    'author': "Na'aman Hirschfeld",
    'author_email': 'nhirschfeld@gmail.com',
    'maintainer': "Na'aman Hirschfeld",
    'maintainer_email': 'nhirschfeld@gmail.com',
    'url': 'https://github.com/starlite-api/jwt-auth',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from typing import TYPE_CHECKING, List, Optional, Union

from pydantic import BaseModel
from starlite import (
    AbstractAuthenticationMiddleware,
    AuthenticationResult,
    NotAuthorizedException,
)
from starlite.connection import ASGIConnection
from typing_extensions import Literal

from starlite_jwt.token import Token

if TYPE_CHECKING:  # pragma: no cover
    from typing import Any, Awaitable, Callable

    from starlite.types import ASGIApp


class JWTAuthenticationMiddleware(AbstractAuthenticationMiddleware):
    def __init__(
        self,
        app: "ASGIApp",
        exclude: Optional[Union[str, List[str]]],
        algorithm: str,
        auth_header: str,
        retrieve_user_handler: "Callable[[str], Awaitable[Any]]",
        token_secret: str,
    ):
        """This Class is a Starlite compatible JWT authentication middleware.

        It checks incoming requests for an encoded token in the auth header specified,
        and if present retrieves the user from persistence using the provided function.

        Args:
            app: An ASGIApp, this value is the next ASGI handler to call in the middleware stack.
            retrieve_user_handler: A function that receives an instance of 'Token' and returns a user, which can be
                any arbitrary value.
            token_secret: Secret for decoding the JWT token. This value should be equivalent to the secret used to encode it.
            auth_header: Request header key from which to retrieve the token. E.g. 'Authorization' or 'X-Api-Key'.
            algorithm: JWT hashing algorithm to use.
            exclude: A pattern or list of patterns to skip.
        """
        super().__init__(app=app, exclude=exclude)
        self.algorithm = algorithm
        self.auth_header = auth_header
        self.retrieve_user_handler = retrieve_user_handler
        self.token_secret = token_secret

    async def authenticate_request(self, connection: "ASGIConnection[Any,Any,Any]") -> AuthenticationResult:
        """Given an HTTP Connection, parse the JWT api key stored in the header
        and retrieve the user correlating to the token from the DB.

        Args:
            connection: An Starlette HTTPConnection instance.

        Returns:
            AuthenticationResult

        Raises:
            [NotAuthorizedException][starlite.exceptions.NotAuthorizedException]: If token is invalid or user is not found.
        """
        encoded_token = connection.headers.get(self.auth_header)

        if not encoded_token:
            raise NotAuthorizedException("No JWT token found in request header")
        return await self.authenticate_token(encoded_token=encoded_token)

    async def authenticate_token(self, encoded_token: "Any") -> AuthenticationResult:
        """Given an encoded JWT token, parse, validate and look up sub within
        token.

        Args:
            encoded_token (Any): _description_

        Raises:
            [NotAuthorizedException][starlite.exceptions.NotAuthorizedException]: If token is invalid or user is not found.

        Returns:
            AuthenticationResult: _description_
        """
        token = Token.decode(
            encoded_token=encoded_token,
            secret=self.token_secret,
            algorithm=self.algorithm,
        )
        user = await self.retrieve_user_handler(token.sub)

        if not user:
            raise NotAuthorizedException()

        return AuthenticationResult(user=user, auth=token)


class CookieOptions(BaseModel):
    path: str = "/"
    """Path fragment that must exist in the request url for the cookie to be valid. Defaults to '/'."""
    domain: Optional[str] = None
    """Domain for which the cookie is valid."""
    secure: bool = False
    """Https is required for the cookie."""
    samesite: Literal["lax", "strict", "none"] = "lax"
    """Controls whether or not a cookie is sent with cross-site requests. Defaults to 'lax'."""
    description: Optional[str] = None
    """Description of the response cookie header for OpenAPI documentation"""


class JWTCookieAuthenticationMiddleware(JWTAuthenticationMiddleware):
    def __init__(
        self,
        app: "ASGIApp",
        exclude: Optional[Union[str, List[str]]],
        algorithm: str,
        auth_header: str,
        auth_cookie: str,
        auth_cookie_options: CookieOptions,
        retrieve_user_handler: "Callable[[str], Awaitable[Any]]",
        token_secret: str,
    ):
        """This Class is a Starlite compatible JWT authentication middleware
        with cookie support.

        It checks incoming requests for an encoded token in the auth header or cookie name specified,
        and if present retrieves the user from persistence using the provided function.

        Args:
            app: An ASGIApp, this value is the next ASGI handler to call in the middleware stack.
            retrieve_user_handler: A function that receives an instance of 'Token' and returns a user, which can be
                any arbitrary value.
            token_secret: Secret for decoding the JWT token. This value should be equivalent to the secret used to encode it.
            auth_header: Request header key from which to retrieve the token. E.g. 'Authorization' or 'X-Api-Key'.
            auth_cookie: Cookie name from which to retrieve the token. E.g. 'token' or 'accessToken'.
            auth_cookie_options:   Cookie configuration options to use when creating cookies for requests.
            algorithm: JWT hashing algorithm to use.
            exclude: A pattern or list of patterns to skip.
        """
        super().__init__(
            algorithm=algorithm,
            app=app,
            auth_header=auth_header,
            retrieve_user_handler=retrieve_user_handler,
            token_secret=token_secret,
            exclude=exclude,
        )

        self.auth_cookie = auth_cookie
        self.auth_cookie_options = auth_cookie_options or CookieOptions()

    async def authenticate_request(self, connection: "ASGIConnection[Any,Any,Any]") -> AuthenticationResult:
        """Given an HTTP Connection, parse the JWT api key stored in the header
        and retrieve the user correlating to the token from the DB.

        Args:
            connection: An Starlette HTTPConnection instance.

        Returns:
            AuthenticationResult

        Raises:
            [NotAuthorizedException][starlite.exceptions.NotAuthorizedException]: If token is invalid or user is not found.
        """
        encoded_token = connection.headers.get(self.auth_header) or connection.cookies.get(self.auth_cookie)
        if not encoded_token:
            raise NotAuthorizedException("No JWT token found in request header or cookies")

        return await self.authenticate_token(encoded_token=encoded_token)
